// Marcelo Eiji Morioka
// Marcos Tatsuo Yamamoto
// Geometria Computacional
// Prof. Coelho
// Linguagem utilizada: Java (jdk 1.1.6)

/* AVLTree.java */

import java.util.Stack;
import java.util.BitSet;

/**
 * AVLTree class.  Implements a height balanced binary tree data structure.
 * Elements contain a referencing key and a data value.  Any object may be used
 * as a key and/or data member.
 *
 * @version:  1.5, 04/26/96 
 * @author:   Paul Brown
 */
public class AVLTree {
    /**
     * The root node of the tree.
     */
    private AVLNode root;
    
    /**
     * The total number of nodes in the tree.
     */
    private int node_count;
    
    /**
     * Constructs a new empty tree.
     */
    public AVLTree() {
	node_count = 0;
    }
    
    /**
     * Returns true if the tree contains no elements.
     */
    public boolean empty() {
	return(root == null);
    }
    
    /**
     * Puts the specified element into the tree.  The element is indexed by
     * the hashcode of the specified key.  The same element may be retrieved
     * by calling get() with the same key.  If an element with the same key
     * already exists within the tree, it is replaced by the new element and
     * the data value of the discarded element is returned.
     * @param key the key used to index the element
     * @param data the specified element's contents
     * @return the replaced element's data value, or null if the tree did
     * not previously contain an element with the specified key
     * @see java.lang.Object#hashCode
     * @see AVLTree#get
     */
    /* flag 0 -> Segmento
       flag 1 -> Ponto*/
    public Object put(int key, Object data, int flag, int x) {
	AVLNode existing_node;
	Object old_data = null;
	if (root == null)
	    root = new AVLNode(key, data);
	else {
	    existing_node = new AVLNode();
	    root = root.put(key, data, existing_node, flag, x);
	    old_data = existing_node.data;
	}
	if (old_data == null)
	    node_count++;
	return(old_data);
    }
    
    /**
     * Gets the element associated with the specified key.
     * @param key the key of the element to get
     * @return the data value of the element referenced by the key, or null
     * if the key was not found
     * @see AVLTree#put
     */
    public Object get(int key) {
	if (root == null)
	    return(null);
	else
	    return(root.get(key));
    }
    
    /**
     * Removes the element indexed by the specified key.  Does nothing if
     * the key is not present.  If an element is removed from the tree, it's
     * data value is returned as the result of this operation.
     * @param key the key of the element to remove
     * @return the removed element's data value, or null if the tree did not
     * contain an element with the specified key
     */
    public Object remove(int key, Object data, int flag, int x) {
	AVLNode existing_node;
	Object old_data = null;
	if (root != null) {
	    existing_node = new AVLNode();
	    root = root.remove(key, data, existing_node, flag, x);
	    old_data = existing_node.data;
	    if (old_data != null)
		node_count--;
	}
	return(old_data);
    }
    
    public Object min() {
	AVLNode aux;
	Object p;
	
	if (root == null)
	    return null;
	
	for (aux = root; aux.left != null; aux = aux.left);
	
	p = remove(((Ponto)aux.data).getX(), aux.data, 1, 0);
	return p;
    }

    /**
     * Removes all nodes from the tree.
     */
    public void clear() {
	root = null;
	node_count = 0;
    }
    
    /**
     * Returns an array containing key/data pairs of all elements in the
     * tree.
     * @return element array, or null if the tree is empty
     */
    public ValuePair[] elements() {
	ValuePair elements[];
	Stack stack;
	AVLNode subtree = root;
	int element_index = 0;
	if (root == null)
	    return(null);
	else {
	    elements = new ValuePair[node_count];
	    stack = new Stack();
	    // perform an inorder traversal of tree
	    while ((subtree != null) || (stack.empty() == false)) {
		while (subtree != null) {
		    stack.push(subtree);
		    subtree = subtree.left;
		}
		subtree = (AVLNode)stack.pop();
				// add this element to element array
		elements[element_index] = new ValuePair(new Integer(subtree.key), subtree.data);
		element_index++;
		subtree = subtree.right;
	    }
	    return(elements);
	}
    }
    
	/**
	 * Returns a String representation of an instance of this class.
	 */
    public String toString() {
	Stack stack = new Stack();
	BitSet levels = new BitSet();
	AVLNode subtree = root;
	int level = 0, level_offset = 0, zero_count = 0;
	StringBuffer buffer = new StringBuffer();
	ValuePair element_array[] = elements();
	boolean backtracked;
	if (element_array == null)
	    buffer.append("tree is empty\n");
	else {
	    buffer.append("tree (showing key values):\n");
	    while ((subtree != null) || (stack.empty() == false)) {
		if (subtree != null)
		    level_offset++;
		while (subtree != null) {
		    stack.push(subtree);
		    level++;
		    subtree = subtree.right;
		}
		subtree = (AVLNode)stack.pop();
		level--;
		if (level == 0) {
		    level = ++zero_count;
		    level_offset = 0;
		}
		if ((level + level_offset) == 1)
		    buffer.append("root=");
		else {
		    for (int i = 0; i < (level + level_offset - 1); i++) {
			if (levels.get(i) == true)
			    buffer.append("|");
			else
			    buffer.append(" ");
			buffer.append("    ");
		    }
		    buffer.append("+----");
		    if (levels.get(level + level_offset - 1) == true)
			levels.clear(level + level_offset - 1);
		    else
			levels.set(level + level_offset - 1);
		}
		buffer.append("[").append(Integer.toString(subtree.key)).append("]\n");
		subtree = subtree.left;
		if (subtree != null)
		    levels.set(level + level_offset);
		else {
		    levels.clear(level + level_offset);
		    // is current subtree a left subtree?
		    backtracked = false;
		    do
			if (((level + level_offset - 1) > 1)
			    && (levels.get(level + level_offset - 1) == false)
			    && (level_offset > 0))
			    level_offset--;
			else
			    backtracked = true;
		    while (backtracked == false);
		}
		for (int i = 0; i <= (level + level_offset); i++) {
		    if (levels.get(i) == true)
			buffer.append("|");
		    else
			buffer.append(" ");
		    if (i < (level + level_offset))
			buffer.append("    ");
		    else
			buffer.append("\n");
		}
	    }
	    buffer.append("tree elements=[");
	    for (int i = 0; i < element_array.length; i++) {
		buffer.append("{key=").append(element_array[i].key().toString());
				//buffer.append(", data=").append(element_array[i].data().toString());
		buffer.append(", data=").append(((Segmento)(element_array[i].data())).getFirst().getX());
		buffer.append(", data=").append(((Segmento)(element_array[i].data())).getFirst().getY());
		if (i < (element_array.length - 1))
		    buffer.append("}, ");
		else
		    buffer.append("}]\n");
	    }
	}
	return(buffer.toString());
    }
    
    // Metodo que procura o primeiro vizinho acima de s
    public Segmento VizinhoAcima(Segmento s, int x) {
	AVLNode proximo, pai = null;
	int det;

	proximo = root;
	if (proximo == null) return null;
	while ((proximo != null) && proximo.key != s.getKey()) {
	    det = ((Segmento)(proximo.data)).Orientacao(s.retornaPonto(x));
	    if (det < 0) {
		pai = proximo;
		proximo = proximo.left;
	    }		
	    else {
		proximo = proximo.right;
		if (proximo!=null)
		    if ((proximo.key == s.getKey()) && (proximo.right==null))
			break;
	    }
	}
	if  ((proximo != null) && proximo.right != null) {
	    proximo = proximo.right;
	    while ((proximo != null) && (proximo.left != null))
		proximo = proximo.left;
	    return ((Segmento)(proximo.data));
	}
	else 
	    if (pai != null)
		return ((Segmento)(pai.data));
	    else 
		return null;
    }
    
    // metodo que procura o primeiro vizinho abaixo de s
    public Segmento VizinhoAbaixo(Segmento s, int x) {
	AVLNode proximo, pai = null;
	int det;

	proximo = root;
	if (proximo == null) return null;
	while ((proximo != null) && proximo.key != s.getKey()) {
	    det = ((Segmento)(proximo.data)).Orientacao(s.retornaPonto(x));
	    if (det >= 0) {
		pai = proximo;
		proximo = proximo.right;		
	    }		
	    else {
		proximo = proximo.left;
		if (proximo!=null)
		    if ((proximo.key == s.getKey()) && (proximo.left==null))
			break;
	    }
	}
	if ((proximo != null) && proximo.left != null) {
	    proximo = proximo.left;
	    while ((proximo != null) && proximo.right != null)
		proximo = proximo.right;
	    return ((Segmento)(proximo.data));
	}
	else 
	    if (pai != null)
		return ((Segmento)(pai.data));
	    else return null;
    }
    
}

/**
 * AVLNode class.  Objects of this type are manipulated by the AVLTree class.
 */
class AVLNode {
    /**
     * The key which is used to reference this node.
     */
    int key;
    
    /**
     * The element.
     */
    Object data;
    
    /**
     * A balance factor, used when rebalancing the tree.
     */
    byte balance;
    
    /**
     * The children of this node.
     */
    AVLNode left, right;
    
    /**
     * Constructs a new node with key and data values uninitialised.
     */
    protected AVLNode() {
	key = 0;
	data = null;
	balance = 0;
    }
    
    /**
     * Constructs a new node with specified key and data values.
     */
    protected AVLNode(int key, Object data) {
	this.key = key;
	this.data = data;
	balance = 0;
    }
    
    /**
     * Inserts a new node into the tree.  If current node has the same key
     * value as the new node, the current node's data value is retrieved and
     * then replaced by the new node's data value.
     */
    protected AVLNode put(int key, Object data, AVLNode existing_node, int flag, int x) {
	byte old_balance;
	int num;
	//		if (key == this.key) {
	// found existing node, replace data value
	//existing_node.data = this.data;
	//this.data = data;
	//		} else 
	if (flag == 0) {
	    if (((Segmento)(this.data)).Orientacao(((Segmento)(data)).retornaPonto(x)) < 0) {
		// insert into left subtree
		if (left != null) {
		    old_balance = left.balance;
		    left = left.put(key, data, existing_node, flag, x);
				// has tree grown?
		    if ((left.balance != old_balance) && (left.balance != 0))
			balance--;
		} else {
		    left = new AVLNode(key, data);
		    balance--;
		}
	    } else {
		// insert into right subtree
		if (right != null) {
		    old_balance = right.balance;
		    right = right.put(key, data, existing_node, flag, x);
				// has tree grown?
		    if ((right.balance != old_balance) && (right.balance != 0))
			balance++;
		} else {
		    right = new AVLNode(key, data);
		    balance++;
		}
	    }
	}
	else {
	    if (((Ponto)(this.data)).getX() > ((Ponto)(data)).getX()) {
		// insert into left subtree
		if (left != null) {
		    old_balance = left.balance;
		    left = left.put(key, data, existing_node, flag, x);
				// has tree grown?
		    if ((left.balance != old_balance) && (left.balance != 0))
			balance--;
		} else {
		    left = new AVLNode(key, data);
		    balance--;
		}
	    } else {
		if (((Ponto)(this.data)).getX() < ((Ponto)(data)).getX()) {
		    // insert into right subtree
		    if (right != null) {
			old_balance = right.balance;
			right = right.put(key, data, existing_node, flag, x);
				// has tree grown?
			if ((right.balance != old_balance) && (right.balance != 0))
			    balance++;
		    } else {
			right = new AVLNode(key, data);
			balance++;
		    }
		}
	    }
	}
	
	// check if we are now out of balance, if so then rebalance
	if ((balance < -1) || (balance > 1))
	    return(rebalanceNode());
	else
	    return(this);
    }
    
    /**
     * A combination of rotations is performed to rebalance the tree based
     * on the balance factor of each node.
     */
    protected AVLNode rebalanceNode() {
	// balance tree rooted at node using rotations
	if (balance < 0)
	    if (left.balance <= 0)
		return(singleRotateRight());
	    else {
				// perform double rotation
		left = left.singleRotateLeft();
		return(singleRotateRight());
	    }
	else
	    if (right.balance >= 0)
		return(singleRotateLeft());
	    else {
				// perform double rotation
		right = right.singleRotateRight();
		return(singleRotateLeft());
	    }
    }
    
    /**
     * Rebalances a right heavy subtree.
     */
    protected AVLNode singleRotateLeft() {
	// perform a single left rotation of the current node
	byte child_balance;
	AVLNode right_child = right;
	// make connections
	right = right_child.left;
	right_child.left = this;
	// update balance factors
	child_balance = right_child.balance;
	if (child_balance <= 0) {
	    if (balance >= 1)
		right_child.balance = (byte)(child_balance - 1);
	    else
		right_child.balance = (byte)(balance + child_balance - 2);
	    balance = (byte)(balance - 1);
	} else {
	    if (balance <= child_balance)
		right_child.balance = (byte)(balance - 2);
	    else
		right_child.balance = (byte)(child_balance - 1);
	    balance = (byte)(balance - child_balance - 1);
	}
	return(right_child);
    }
    
    /**
     * Rebalances a left heavy subtree.
     */
    protected AVLNode singleRotateRight() {
	// perform a single right rotation of the current node
	byte child_balance;
	AVLNode left_child = left;
	// make connections
	left = left_child.right;
	left_child.right = this;
	// update balance factors
	child_balance = left_child.balance;
	if (child_balance <= 0) {
	    if (child_balance > balance)
		left_child.balance = (byte)(child_balance + 1);
	    else
		left_child.balance = (byte)(balance + 2);
	    balance = (byte)(1 + balance - child_balance);
	} else {
	    if (balance <= -1)
		left_child.balance = (byte)(child_balance + 1);
	    else
		left_child.balance = (byte)(balance + child_balance + 2);
	    balance = (byte)(1 + balance);
	}
	return(left_child);
    }
    
    /**
     * Retrieves the data value of the node with the specified key.  If the
     * node is not found, this method returns null.
     */
    protected Object get(int key) {
	if (key == this.key) {
	    // found node
	    return(data);
	} else if (key < this.key) {
	    // search left subtree
	    if (left != null)
		return(left.get(key));
	    else
		return(null);
	} else {
	    // search right subtree
	    if (right != null)
		return(right.get(key));
	    else
		return(null);
	}
    }
    
    /**
     * Removes the node with the specified key value from the tree.
     */
    protected AVLNode remove(int key, Object data, AVLNode existing_node, int flag, int x) {
	byte old_balance;
	if (flag == 0) {
	    if (key == this.key) {
		// found node
		existing_node.data = data;
		if (right == null)
		    return(left);
		else {
				// find and remove leftmost child of right child
		    old_balance = right.balance;
		    AVLNode new_root = new AVLNode(this.key, this.data);
		    right = right.removeLeftmostDescendant(new_root);
				// connect new root
		    new_root.left = left;
		    new_root.right = right;
		    new_root.balance = balance;
		    return(new_root.restoreRightBalance(old_balance));
		}
	    } else 
		if (((Segmento)(this.data)).Orientacao(((Segmento)(data)).retornaPonto(x)) < 0) {			// remove from left child
		    if (left == null) {
				// no left child, key not found
			return(this);
		    } else {
				// do the deletion
			old_balance = left.balance;
			left = left.remove(key, data, existing_node, flag, x);
			return(restoreLeftBalance(old_balance));
		    }
		} else {
		    // remove from right child
		    if (right == null) {
				// no right child, key not found
			return(this);
		    } else {
				// do the deletion
			old_balance = right.balance;
			right = right.remove(key, data, existing_node, flag, x);
			return(restoreRightBalance(old_balance));
		    }
		}
	}
	else {
	    if (key == this.key) {
		// found node
		existing_node.data = data;
		if (right == null)
		    return(left);
		else {
				// find and remove leftmost child of right child
		    old_balance = right.balance;
		    AVLNode new_root = new AVLNode(this.key, this.data);
		    right = right.removeLeftmostDescendant(new_root);
				// connect new root
		    new_root.left = left;
		    new_root.right = right;
		    new_root.balance = balance;
		    return(new_root.restoreRightBalance(old_balance));
		}
	    } else if (key < this.key) {
		// remove from left child
		if (left == null) {
				// no left child, key not found
		    return(this);
		} else {
				// do the deletion
		    old_balance = left.balance;
		    left = left.remove(key, data, existing_node, flag, x);
		    return(restoreLeftBalance(old_balance));
		}
	    } else {
		// remove from right child
		if (right == null) {
				// no right child, key not found
		    return(this);
		} else {
				// do the deletion
		    old_balance = right.balance;
		    right = right.remove(key, data, existing_node, flag, x);
		    return(restoreRightBalance(old_balance));
		}
	    }
	}
    }
    
    /**
     * Finds and retrieves the leftmost descendant of the current node.
     */
    protected AVLNode removeLeftmostDescendant(AVLNode child) {
	byte old_balance;
	if (left == null) {
	    // this is the leftmost descendant
	    child.key = key;
	    child.data = data;
	    // remove self from tree
	    return(right);
	} else {
	    old_balance = left.balance;
	    // do the deletion
	    left = left.removeLeftmostDescendant(child);
	    // fix balance factors
	    return(restoreLeftBalance(old_balance));
	}
    }
    
    /**
     * Restores balance of the left subtree after a node removal.
     */
    protected AVLNode restoreLeftBalance(byte old_balance) {
	// fix balance factors
	if (left == null)
	    balance++;
	else if ((left.balance != old_balance) && (left.balance == 0))
	    balance++;
	// rebalance if necessary
	if (balance > 1)
	    return(rebalanceNode());
	else
	    return(this);
    }
    
    /**
     * Restores balance of the right subtree after a node removal.
     */
    protected AVLNode restoreRightBalance(byte old_balance) {
	// fix balance factors
	if (right == null)
	    balance--;
	else if ((right.balance != old_balance) && (right.balance == 0))
	    balance--;
	// rebalance if necessary
	if (balance < -1)
	    return(rebalanceNode());
	else
	    return(this);
    }
}
